//
//  Playfield.m
//  Demine
//
//  Created by Michael Heyeck on 4/9/10.
//  Copyright Fair Oaks Labs, Inc. 2010. All rights reserved.
//

#import <QuartzCore/QuartzCore.h>
#import "Path.h"
#import "Text.h"
#import "Playfield.h"


const static NSInteger TAG_BASE = 501;

const static CGSize tileSize = { 64,  64};
const static CGSize cellSize = { 64,  64};


@interface Playfield ()

@property (nonatomic, retain) UIScrollView* scrollView;
@property (nonatomic, retain, readonly) NSArray* box;
@property (nonatomic, retain, readonly) NSArray* glyphs;
@property (nonatomic, retain, readonly) NSArray* contents;

- (BOOL)cellForPoint:(CGPoint*)point;
- (UIImage*)imageForCell:(CGPoint)pos;

- (void)removeTiles;
- (void)createTiles;
- (UIView*)addTileForFrame:(CGRect)frame;
- (void)sizeContent;
- (void)reload;

@end


@implementation Playfield

@synthesize content;

- (NSArray*)box
{
	if (!box)
	{
		// Basic (top) trapezoid
		CGFloat w = 5;
		CGPoint edge[4]  = {CGPointMake(0, 0), CGPointMake(cellSize.width, 0), CGPointMake(cellSize.width-w, w), CGPointMake(w, w)};
		CGPoint floor[4]  = {CGPointMake(w, w), CGPointMake(cellSize.width-w, w), CGPointMake(cellSize.width-w, cellSize.width-w), CGPointMake(w, cellSize.width-w)};

		// Paths
		CGMutablePathRef p;
		CGAffineTransform m;
		// Floor
		m = CGAffineTransformMakeScale(1.0, 1.0);
		Path* f = [[Path new] autorelease];
		p = CGPathCreateMutable();
		CGPathAddLines(p, &m, floor, 4);
		f.path = p;
		f.color = [UIColor grayColor];
		f.drawMode = kCGPathFill;
		CGPathRelease(p);
		// Top
		m = CGAffineTransformMakeScale(1.0, 1.0);
		Path* t = [[Path new] autorelease];
		p = CGPathCreateMutable();
		CGPathAddLines(p, &m, edge, 4);
		t.path = p;
		t.color = [UIColor colorWithWhite:.2 alpha:1.0];
		t.drawMode = kCGPathFill;
		CGPathRelease(p);
		// Bottom
		m = CGAffineTransformConcat(CGAffineTransformMakeScale(1.0, -1.0),
									CGAffineTransformMakeTranslation(0, cellSize.height));
		Path* b = [[Path new] autorelease];
		p = CGPathCreateMutable();
		CGPathAddLines(p, &m, edge, 4);
		b.path = p;
		b.color = [UIColor colorWithWhite:.7 alpha:1.0];
		b.drawMode = kCGPathFill;
		CGPathRelease(p);
		// Left
		m = CGAffineTransformConcat(CGAffineTransformMakeRotation(M_PI/2),
									CGAffineTransformMakeScale(-1.0, 1.0));
		Path* l = [[Path new] autorelease];
		p = CGPathCreateMutable();
		CGPathAddLines(p, &m, edge, 4);
		l.path = p;
		l.color = [UIColor colorWithWhite:.1 alpha:1.0];
		l.drawMode = kCGPathFill;
		CGPathRelease(p);
		// Right
		m = CGAffineTransformConcat(CGAffineTransformMakeRotation(M_PI/2),
									CGAffineTransformMakeTranslation(cellSize.width, 0));
		Path* r = [[Path new] autorelease];
		p = CGPathCreateMutable();
		CGPathAddLines(p, &m, edge, 4);
		r.path = p;
		r.color = [UIColor colorWithWhite:.6 alpha:1.0];
		r.drawMode = kCGPathFill;
		CGPathRelease(p);
		
		box = [[NSArray alloc] initWithObjects:f,t,b,l,r,nil];
	}
	return box;
}

- (NSArray*)glyphs
{
	if (!glyphs)
	{
		glyphs = [[NSMutableArray alloc] initWithCapacity:11];
		NSMutableArray* myGlyphs = (NSMutableArray*) glyphs;

		Text* text = [[Text new] autorelease];
		text.drawMode = kCGTextFillStroke;
		text.strokeWidth = 1.0;
		text.font = [UIFont systemFontOfSize:18];
		text.color = [UIColor blackColor];
		text.offset = CGPointMake(.5*cellSize.width-7, .5*cellSize.height-10);
		text.string = @"";
		
		// Numbers
		for (int i = 0; i <= 8; i++)
		{
			NSMutableArray* a = [[[NSMutableArray alloc] initWithArray:self.box copyItems:YES] autorelease];
			
			Text* t = [[text copy] autorelease];
			t.string = [NSString stringWithFormat:@"%d",i];
			[a addObject:t];
			
			[myGlyphs addObject:a];
		}
		
		// Mine
		{
			NSMutableArray* a = [[[NSMutableArray alloc] initWithArray:self.box copyItems:YES] autorelease];
			[[a objectAtIndex:0] setColor:[UIColor redColor]];
			
			Text* t = [[text copy] autorelease];
			t.string = @"X";
			[a addObject:t];
			
			[myGlyphs addObject:a];
		}
		
		// Hidden
		{
			NSMutableArray* a = [[[NSMutableArray alloc] initWithArray:self.box copyItems:YES] autorelease];
			[[a objectAtIndex:1] setColor:[[self.box objectAtIndex:2] color]];
			[[a objectAtIndex:2] setColor:[[self.box objectAtIndex:1] color]];
			[[a objectAtIndex:3] setColor:[[self.box objectAtIndex:4] color]];
			[[a objectAtIndex:4] setColor:[[self.box objectAtIndex:3] color]];
			
			Text* t = [[text copy] autorelease];
			[a addObject:t];
			
			[myGlyphs addObject:a];
		}
	}
	return glyphs;
}

- (NSArray*)contents
{
	if (!contents)
	{
		contents = [[NSMutableArray alloc] initWithCapacity:[self.glyphs count]];
		NSMutableArray* myContents = (NSMutableArray*) contents;
		
		for (NSArray* symbol in self.glyphs)
		{
			// Create and get a pointer to context
			UIGraphicsBeginImageContext(cellSize);
			CGContextRef context = UIGraphicsGetCurrentContext();
			
			// Flop text rightside-up
			CGContextSetTextMatrix(context, CGAffineTransformMakeScale(1, -1));
			
			// Render box
			for (Glyph* g in symbol)
			{
				[g renderToContext:context];
			}
			
			// Create and store image
			[myContents addObject:UIGraphicsGetImageFromCurrentImageContext()];
			
			// Release image context, and continue
			UIGraphicsEndImageContext();
		}
	}
	return contents;
}

- (UIScrollView*)scrollView
{
	return (UIScrollView*) self.view;
}

- (void)setScrollView:(UIScrollView*)newScrollView
{
	self.view = newScrollView;
}


// The parent's designated initializer. Overriden to call the new DI.
- (id)initWithNibName:(NSString*)nibNameOrNil bundle:(NSBundle*)nibBundleOrNil
{
	return [self initWithNibName:nibNameOrNil bundle:nibBundleOrNil puzzle:nil];
}


// The new designated initializer.
- (id)initWithNibName:(NSString*)nibNameOrNil bundle:(NSBundle*)nibBundleOrNil puzzle:(Puzzle*)ipuzzle
{
    if (self = [super initWithNibName:nibNameOrNil bundle:nibBundleOrNil])
	{
		if (ipuzzle)
		{
			NSLog(@"%@", ipuzzle);
			puzzle = [ipuzzle retain];
		}
		else
		{
			[self release];
			return nil;
		}
    }
    return self;
}


/*
// Implement loadView to create a view hierarchy programmatically, without using a nib.
- (void)loadView {
}
*/


// Implement viewDidLoad to do additional setup after loading the view, typically from a nib.
- (void)viewDidLoad
{
    [super viewDidLoad];

	world = CGSizeMake(cellSize.width*puzzle.w, cellSize.height*puzzle.h);
	scale = 1.0;
	
	[self createTiles];
	tileBox = CGRectZero;
	
	[self sizeContent];
	self.scrollView.contentOffset = CGPointMake(-.05*self.scrollView.frame.size.width,
												-.05*self.scrollView.frame.size.height);
	[self.scrollView flashScrollIndicators];
}


/*
// Override to allow orientations other than the default portrait orientation.
- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation {
    // Return YES for supported orientations
    return (interfaceOrientation == UIInterfaceOrientationPortrait);
}
*/

- (void)didReceiveMemoryWarning {
	// Releases the view if it doesn't have a superview.
    [super didReceiveMemoryWarning];
	
	// Release any cached data, images, etc that aren't in use.
}

- (void)viewDidUnload {
	// Release any retained subviews of the main view.
	// e.g. self.myOutlet = nil;
}


- (void)dealloc
{
	// Clear weak delegate reference
	if (self.scrollView.delegate == self)
	{
		self.scrollView.delegate = nil;
	}
	
	[box release];
	[glyphs release];
	[contents release];
	[puzzle release];
	[self removeTiles];
	[content release];
    [super dealloc];
}

#pragma mark Scroll view delegate methods

- (void)scrollViewDidScroll:(UIScrollView*)scrollView
{
	// Tile width and height (in view space)
	CGFloat w = tileSize.width;
	CGFloat h = tileSize.height;
	
	// Legal region (in view space)
	CGRect legal = CGRectMake(0, 0, world.width*scale, world.height*scale);
	
	// Displayed region (in view space)
	CGRect viewport = CGRectIntersection([self.content convertRect:self.scrollView.bounds fromView:self.scrollView], legal);
	
	// For existing tiles
	for (UIView* tv in [self.content subviews])
	{
		// If there's no overlap with viewport, toss the tile
		if (!CGRectIntersectsRect(viewport, tv.frame))
		{
			[tv removeFromSuperview];
			[extraTiles addObject:tv];
		}
	}
	
	NSInteger oldMinCol = tileBox.origin.x;
	NSInteger oldMinRow = tileBox.origin.y;
	NSInteger oldMaxCol = CGRectGetMaxX(tileBox)-1;
	NSInteger oldMaxRow = CGRectGetMaxY(tileBox)-1;
	
	// For new tiles
	tileBox = CGRectIntegral(CGRectApplyAffineTransform(viewport, CGAffineTransformMakeScale(1/w, 1/h)));
	for (int bw = tileBox.size.width, n = bw*tileBox.size.height-1; n >= 0; n--)
	{
		NSInteger c = tileBox.origin.x + n % bw;
		NSInteger r = tileBox.origin.y + n / bw;
		
		//	If missing, create
		if ((c < oldMinCol) || (c > oldMaxCol) || (r < oldMinRow) || (r > oldMaxRow))
		{
			UIView* tv = [self addTileForFrame:CGRectIntersection(CGRectMake(c*w, r*h, w, h), legal)];
			tv.layer.contents = (id) [[self imageForCell:CGPointMake(c, r)] CGImage];
			tv.tag = TAG_BASE + puzzle.w*r + c;
		}
	}
}

#pragma mark Tap detecting view delegate methods

- (void)tapDetectingView:(TapDetectingView*)tdView gotSingleTapAtPoint:(CGPoint)tapPoint
{
	if (![self cellForPoint:&tapPoint]) return;
	NSLog(@"Tap at (%.1f, %.1f)", tapPoint.x, tapPoint.y);

	[puzzle revealCell:tapPoint];

	UIView* tv = [content viewWithTag:(TAG_BASE + puzzle.w*tapPoint.y + tapPoint.x)];
	tv.layer.contents = (id) [[self imageForCell:tapPoint] CGImage];
}


- (void)tapDetectingView:(TapDetectingView*)tdView gotDoubleTapAtPoint:(CGPoint)tapPoint
{
	[self tapDetectingView:tdView gotSingleTapAtPoint:tapPoint];
}

#pragma mark Extension methods

- (BOOL)cellForPoint:(CGPoint*)point
{
	if (!point) return FALSE;
	
	CGPoint cell = CGPointMake(floorf(point->x/(cellSize.width*scale)),
							   floorf(point->y/(cellSize.height*scale)));
	
	if ([puzzle isLegalCell:cell])
	{
		*point = cell;
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}


- (UIImage*)imageForCell:(CGPoint)pos
{
	if (![puzzle isRevealedCell:pos])
	{
		return [self.contents objectAtIndex:10];
	}
	else if ([puzzle isMinedCell:pos])
	{
		return [self.contents objectAtIndex:9];
	}
	else
	{
		return [self.contents objectAtIndex:[puzzle neighborsOfCell:pos]];
	}
}


- (void)removeTiles
{
	for (UIView* tv in tiles)
	{
		// Remove tiles from content view
		[tv removeFromSuperview];
	}
	[tiles release];
	tiles = nil;
	
	[extraTiles release];
	extraTiles = nil;
}


- (void)createTiles
{
	if (tiles) [self removeTiles];
	
	tiles = [[NSMutableArray alloc] initWithCapacity:4];
	extraTiles = [[NSMutableSet alloc] initWithCapacity:4];
}


- (UIView*)addTileForFrame:(CGRect)frame
{
	UIView* tv = [extraTiles anyObject];
	if (tv)
	{
		// tv also retained by tiles
		[extraTiles removeObject:tv];
	}
	else
	{
		tv = [[UIView new] autorelease];
		tv.clearsContextBeforeDrawing = NO;
		tv.backgroundColor = [UIColor grayColor];
		[tiles addObject:tv];
	}
	
	[self.content addSubview:tv];
	tv.frame = frame;
	[tv setNeedsDisplay];
	return tv;
}


- (void)sizeContent
{
	self.content.frame = CGRectMake(0, 0, world.width*scale, world.height*scale);
	self.scrollView.contentSize = CGSizeMake(CGRectGetMaxX(self.content.frame),
											 CGRectGetMaxY(self.content.frame));
}


- (void)reload
{
	// Recycle all tiles
	for (UIView* tv in [self.content subviews])
	{
		[tv removeFromSuperview];
		[extraTiles addObject:tv];
	}
	tileBox = CGRectZero;
	
	// Trigger re-tiling
	[self scrollViewDidScroll:self.scrollView];
}

@end
